package fr.jelora;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.OpenOption;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Random;

import javax.xml.crypto.dsig.spec.ExcC14NParameterSpec;

import jssc.SerialPort;
import jssc.SerialPortException;
import jssc.SerialPortTimeoutException;

public class TestOrdiZ80CircuitDev {

	public static void main(String[] args) {
		TestOrdiZ80CircuitDev testOrdiZ80CircuitDev = new TestOrdiZ80CircuitDev();
		testOrdiZ80CircuitDev.run();
	}
	
	public void run(){
		try{
			byte[] programmeTest = new byte[]{
					// LD A,(1)
					(byte) 0b00111110,
					(byte) 0b00000001,
				// Debut
					// OUT (0),A
					(byte) 0b11010011,
					(byte) 0b00000000,
					// RLCA
					(byte) 0b00000111,
					// BIT (7),A
					(byte) 0b11001011,
					(byte) 0b01111111,
					// JR NZ,(3)
					(byte) 0b00100000,
					(byte) 0b00000011,
					// JP Debut
					(byte) 0b11000011,
					(byte) 0b00000010,
					(byte) 0b00000000,
				// Debut2
					// OUT (0),A
					(byte) 0b11010011,
					(byte) 0b00000000,
					// RRCA
					(byte) 0b00001111,
					// BIT (0),A
					(byte) 0b11001011,
					(byte) 0b01000111,
					// JRN Z,(3)
					(byte) 0b00100000,
					(byte) 0b00000011,
					// JP Debut2
					(byte) 0b11000011,
					(byte) 0b00001100,
					(byte) 0b00000000,
					// JP Debut
					(byte) 0b11000011,
					(byte) 0b00000010,
					(byte) 0b00000000
			};
			
			SerialPort serial = new SerialPort("COM5");
			serial.openPort();
			//serial.setParams(SerialPort.BAUDRATE_19200, SerialPort.DATABITS_8, SerialPort.STOPBITS_1, SerialPort.PARITY_NONE);
			//serial.setParams(SerialPort.BAUDRATE_57600, SerialPort.DATABITS_8, SerialPort.STOPBITS_1, SerialPort.PARITY_NONE);
			serial.setParams(230400, SerialPort.DATABITS_8, SerialPort.STOPBITS_1, SerialPort.PARITY_NONE);
			
			// Vérification de la liaison
			boolean liaisonOk = this.testLiaisonOk(serial);
			System.out.print("Liaison : ");
			if ( this.testLiaisonOk(serial) ){
				System.out.println("OK\n");
			}
			else{
				System.out.println("ERR");
				throw new Exception();
			}
			
			// Vérification du type de circuit de développement
			System.out.print("Type circuit de développement : ");
			if ( this.testInfoCircuitDeDeveloppement(serial) ){
				System.out.println("OK\n");
			}
			else{
				System.out.println("ERR");
				throw new Exception();
			}
			
			// Début séquence de programmation
			System.out.print("Début séquence de programmation : ");
			if ( this.startDev(serial) ){
				System.out.println("OK\n");
			}
			else{
				System.out.println("ERR");
				throw new Exception();
			}
			
			// Définition du temps d'écriture ROM
			int tempsEcriture = 400;
			System.out.print("Définition du temps d'écriture à " + tempsEcriture + " : ");
			if ( this.definitionTempsEcriture(serial, tempsEcriture) ){
				System.out.println("OK\n");
			}
			else{
				System.out.println("ERR");
				throw new Exception();
			}
			
			// Effacement ROM
			/*System.out.println("Effacement ROM :");
			if ( this.ecritureROM(serial, new byte[1024]) ){
				System.out.println("OK\n");
			}
			else{
				System.out.println("ERR");
				throw new Exception();
			}*/
			
			// Ecriture du programme de test
			System.out.println("Ecriture du programme de test :");
			if ( this.ecritureROM(serial, programmeTest) ){
				System.out.println("OK\n");
			}
			else{
				System.out.println("ERR");
				throw new Exception();
			}
			
			// Ecriture d'un fichier en ROM
			/*System.out.println("Ecriture d'un fichier en ROM :");
			File fichierTest = new File("fichier_test2.txt");
			if ( fichierTest.exists() ){
				Date dateDebut = new Date();
				
				byte[] contenuFichier = Files.readAllBytes(fichierTest.toPath());
				if ( this.ecritureROM(serial, programmeTest) ){
					System.out.println("OK\n");
				}
				else{
					System.out.println("ERR");
					throw new Exception();
				}
				
				Date dateFin = new Date();
				long tempsDeProg = (dateFin.getTime() - dateDebut.getTime()) / 1000;
				int minutes = (int) (tempsDeProg / 60);
				int secondes = (int) (tempsDeProg - (minutes*60));
				System.out.println("Temps de programmation : " + minutes + " min " + secondes);
			}
			else{
				System.out.println("Le fichier n'existe pas !");
				throw new Exception();
			}*/
			
			// Ecriture contenu aléatoire en ROM
			/*System.out.println("Ecriture contenu aléatoire en ROM :");
			byte[] contenuAleatoire = new byte[1024];
			Random random = new Random();
			random.nextBytes(contenuAleatoire);
			Date dateDebutEcritureAleatoire = new Date();
			if ( this.ecritureROM(serial, contenuAleatoire) ){
				System.out.println("OK");
				Date dateFin = new Date();
				long tempsDeProg = (dateFin.getTime() - dateDebutEcritureAleatoire.getTime()) / 1000;
				int minutes = (int) (tempsDeProg / 60);
				int secondes = (int) (tempsDeProg - (minutes*60));
				System.out.println("Temps de programmation : " + minutes + " min " + secondes + "\n");
			}
			else{
				System.out.println("ERR");
				throw new Exception();
			}*/
			
			// Lecture mémoire
			/*System.out.println("Lecture ROM :");
			byte[] contenuROM = this.lectureROM(serial, 0, 63);
			if ( contenuROM != null ){
				for ( int i = 6 ; i < 6+5 ; i++ ){
					System.out.println(" - 0x" + this.bytesToHex(new byte[]{ (byte) (i-6) }) + " = " + this.byteToBitString(contenuROM[i]));
				}
				System.out.println(new String(contenuROM).substring(0, 11));
				File fileResult = new File("result.txt");
				FileOutputStream fos = new FileOutputStream(fileResult);
				fos.write(contenuROM);
				
				System.out.println();
			}
			else{
				System.out.println("ERR");
				throw new Exception();
			}*/
			
			// Lecture mémoire du contenu aléatoire envoyé
			/*System.out.println("Comparaison ROM :");
			byte[] contenuROM = this.lectureROM(serial, 0, contenuAleatoire.length - 1);
			if ( contenuROM != null ){
				for ( int i = 0 ; i < contenuAleatoire.length ; i++ ){
					if ( contenuROM[i] != contenuAleatoire[i] ){
						System.out.println(" - 0x" + this.bytesToHex(new byte[]{ (byte) (i) }) + " = " + this.byteToBitString(contenuROM[i]) + " =/= " + this.byteToBitString(contenuAleatoire[i]));
					}
				}
				System.out.println("OK\n");
			}
			else{
				System.out.println("ERR");
				throw new Exception();
			}*/
			
			
			// Fin séquence de programmation
			System.out.print("Fin séquence de programmation : ");
			if ( this.endDev(serial) ){
				System.out.println("OK");
			}
			else{
				System.out.println("ERR");
			}
			
			serial.closePort();
		}
		catch(Exception e){
			e.printStackTrace();
		}
	}
	
	private boolean definitionTempsEcriture(SerialPort serial, int tempsEcriture) throws SerialPortException, IOException{
		
		int tempsEcritureHigh = tempsEcriture / 256;
		int tempsEcritureLow = tempsEcriture - (tempsEcritureHigh*256);
		
		ByteArrayOutputStream commandeBuilder = new ByteArrayOutputStream();
		commandeBuilder.write("WRITEDELAY".getBytes());
		commandeBuilder.write(new byte[]{ (byte) tempsEcritureHigh, (byte) tempsEcritureLow });
		
		String reponse = this.sendCommandToString(serial, commandeBuilder.toByteArray());
		if ( reponse.equals("OK") ){
			return true;
		}
		else{
			return false;
		}
	}
	
	private byte[] lectureROM(SerialPort serial, int adressFrom, int adressTo) throws SerialPortException, IOException{
		
		int adressFromHigh = adressFrom / 256;
		int adressFromLow = adressFrom - (adressFromHigh*256);
		
		int adressToHigh = adressTo / 256;
		int adressToLow = adressTo - (adressToHigh*256);
		
		ByteArrayOutputStream commandeBuilder = new ByteArrayOutputStream();
		commandeBuilder.write("READ".getBytes());
		commandeBuilder.write(new byte[]{ (byte) adressFromHigh, (byte) adressFromLow, (byte) adressToHigh, (byte) adressToLow });
		byte[] reponseByteArray = this.sendCommandToByteArray(serial, commandeBuilder.toByteArray(), false, adressTo - adressFrom - 1 + 6);
		if ( new String(reponseByteArray).startsWith("VALUES") && reponseByteArray.length > 6 ){
			byte[] result = new byte[reponseByteArray.length - 6];
			System.arraycopy(reponseByteArray, 6, result, 0, reponseByteArray.length - 6);
			return result;
		}
		else{
			return null;
		}
	}
	
	private boolean ecritureROM(SerialPort serial, byte[] data) throws SerialPortException, IOException{
		System.out.println("Ecriture de " + data.length + " octets");
		ByteArrayOutputStream commandeBuilder;
		byte[] buffer;
		int tailleBloc = 256;
		for ( int i = 0 ; i < data.length ; i += tailleBloc ){
			int adressHigh = i / 256;
			int adressLow = i - (adressHigh*256);
				
			if ( i + tailleBloc < data.length ){
				buffer = new byte[tailleBloc];
				System.arraycopy(data, i, buffer, 0, tailleBloc);
			}
			else{
				buffer = new byte[data.length - i];
				System.arraycopy(data, i, buffer, 0, data.length - i);
			}
				
			System.out.print("WRITE de " + buffer.length + " octets -> ");
			commandeBuilder = new ByteArrayOutputStream();
			commandeBuilder.write("WRITE".getBytes());
			commandeBuilder.write(new byte[]{ (byte) adressHigh, (byte) adressLow, (byte) (buffer.length - 1) });
			String reponse = this.sendCommandToString(serial, commandeBuilder.toByteArray());
			System.out.print(reponse);
			if ( reponse.equals("SEND") ){
				// OK
				reponse = this.sendCommandToString(serial, buffer, true);
				System.out.println(" -> " + reponse);
				if ( !reponse.equals("OK") ){
					return false;
				}
			}
			else{
				System.out.println("ERR");
				return false;
			}
		}
		return true;
	}
	
	private boolean endDev(SerialPort serial) throws SerialPortException{
		String reponse = this.sendCommandToString(serial, "END".getBytes());
		//System.out.print("Fin séquence de programmation : ");
		if ( reponse != null ){
			//System.out.println(reponse);
			if ( reponse.equals("OK") ){
				return true;
			}
			else{
				return false;
			}
		}
		else{
			return false;
		}
	}
	
	private boolean startDev(SerialPort serial) throws SerialPortException{
		String reponse = this.sendCommandToString(serial, "START".getBytes());
		//System.out.print("Début séquence de programmation : ");
		if ( reponse != null ){
			//System.out.println(reponse);
			if ( reponse.equals("OK") ){
				return true;
			}
			else{
				return false;
			}
		}
		else{
			return false;
		}
	}
	
	private boolean testInfoCircuitDeDeveloppement(SerialPort serial) throws SerialPortException{
		String reponse = this.sendCommandToString(serial, "INFO".getBytes());
		if (reponse != null) {
			//System.out.print("\"" + reponse + "\" -> ");
			if (reponse.equals("OrdiZ80-CircuitDev-1")) {
				return true;
			} 
			else {
				return false;
			}
		} 
		else {
			return false;
		}
	}
	
	private boolean testLiaisonOk(SerialPort serial) throws SerialPortException{
		String reponse = this.sendCommandToString(serial, "TEST".getBytes());
		if ( reponse != null ){
			if ( reponse.equals("OK") ){
				// Liaison OK
				//System.out.println("Réponse: \"" + reponse + "\"");
				return true;
			}
			else{
				return false;
			}
		}
		else{
			return false;
		}
	}
	
	private String sendCommandToString(SerialPort serial, byte[] commande) throws SerialPortException {
		return new String(this.sendCommandToByteArray(serial, commande, false, 0));
	}
	
	private String sendCommandToString(SerialPort serial, byte[] commande, boolean sansR) throws SerialPortException {
		return new String(this.sendCommandToByteArray(serial, commande, sansR, 0));
	}
	
	private byte[] sendCommandToByteArray(SerialPort serial, byte[] commande, boolean sansR, int tailleReponseFixe) throws SerialPortException {
		byte[] reponse = null;
		if ( serial != null && commande != null ){
			if ( serial.isOpened() ){
				// Envoi de la commande
				for ( byte byteCommande : commande ){
					serial.writeByte(byteCommande);
					try {
						Thread.sleep(0, 1);
					} 
					catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				if ( !sansR ){
					serial.writeByte((byte) '\r');
					try {
						Thread.sleep(1);
					} 
					catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				
				// Récupération de la réponse (timout 1s)
				byte[] temp = null;
				ByteArrayOutputStream byteArrayBuilder = new ByteArrayOutputStream();
				int timeout = 0;
				while ( true ){
					
					temp = serial.readBytes();
					if ( temp != null ){
						try {
							byteArrayBuilder.write(temp);
						} 
						catch (IOException e) {
							e.printStackTrace();
						}
						
						if ( tailleReponseFixe > 0 && byteArrayBuilder.size() == tailleReponseFixe ){
							break;
						}
						else if ( byteArrayBuilder.toString().endsWith("\n\r") ){
							// Fin de la réponse reçue
							byte[] reponseComplete = byteArrayBuilder.toByteArray();
							if ( reponseComplete.length - 2 > 0 ){
								reponse = new byte[reponseComplete.length - 2];
								System.arraycopy(reponseComplete, 0, reponse, 0, reponseComplete.length - 2);
							}
							break;
						}
						
						timeout = 0;
					}
					else{
						timeout += 10;
					}
					
					if ( timeout >= 60000 ){
						break;
					}
					try {
						Thread.sleep(10);
					} 
					catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}
		}
		return reponse;
	}
	
	private String byteToBitString(byte valeur){
		StringBuilder bitString = new StringBuilder();
		if ( (valeur & 0b10000000) > 0 ){
			bitString.append("1");
		}
		else{
			bitString.append("0");
		}
		if ( (valeur & 0b01000000) > 0 ){
			bitString.append("1");
		}
		else{
			bitString.append("0");
		}
		if ( (valeur & 0b00100000) > 0 ){
			bitString.append("1");
		}
		else{
			bitString.append("0");
		}
		if ( (valeur & 0b00010000) > 0 ){
			bitString.append("1");
		}
		else{
			bitString.append("0");
		}
		if ( (valeur & 0b00001000) > 0 ){
			bitString.append("1");
		}
		else{
			bitString.append("0");
		}
		if ( (valeur & 0b00000100) > 0 ){
			bitString.append("1");
		}
		else{
			bitString.append("0");
		}
		if ( (valeur & 0b00000010) > 0 ){
			bitString.append("1");
		}
		else{
			bitString.append("0");
		}
		if ( (valeur & 0b00000001) > 0 ){
			bitString.append("1");
		}
		else{
			bitString.append("0");
		}
		return bitString.toString();
	}
	
	private String bytesToHex(byte[] data){
		StringBuffer hexString = new StringBuffer();
		for (int i = 0; i < data.length; i++) {
			String hex = Integer.toHexString(0xff & data[i]).toUpperCase();
			if (hex.length() == 1) {
				hexString.append('0');
			}
			hexString.append(hex);
		}
		return hexString.toString();
	}

}
